package com.mindoo.ftp.controller;

import java.io.File;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.util.logging.Level;

import org.apache.ftpserver.FtpServer;
import org.apache.ftpserver.FtpServerFactory;
import org.apache.ftpserver.filesystem.nativefs.NativeFileSystemFactory;

import com.mindoo.ftp.connection.DefaultConnectionConfig;
import com.mindoo.ftp.connection.nsf.NSFDataConnectionConfiguration;
import com.mindoo.ftp.internal.FTPActivator;
import com.mindoo.ftp.listeners.ListenerFactory;
import com.mindoo.ftp.users.nsf.NSFUserManager;

/**
 * Controller class to start, stop and restart the FTP server
 * 
 * @author Karsten Lehmann
 */
public class FTPController {
	private static boolean m_started;
	private static FtpServer m_server;
	private static boolean m_serverDisabled;
	private static int m_port;

	public static synchronized void startServer() throws Exception {
		if (m_started)
			return;

		setupServer();
		if (m_server!=null) {
			AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {

				@Override
				public Object run() throws Exception {
					m_server.start();
					System.out.println("Mindoo FTP running on port "+m_port);
					FTPActivator.log(Level.INFO, "Mindoo FTP running on port "+m_port, null);
					return null;
				}
			});
		}
		m_started=true;
	}

	public static synchronized void stopServer() throws Exception {
		if (m_serverDisabled) {
			m_started=false;
			return;
		}
		else if (!m_started)
			return;

		AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {

			@Override
			public Object run() throws Exception {
				m_server.stop();
				return null;
			}
		});
		m_started=false;
	}

	public static synchronized boolean isRunning() {
		return m_started;
	}

	public static synchronized void dispose() throws Exception {
		stopServer();
		m_server=null;
	}

	public static synchronized void restartServer() throws Exception {
		stopServer();
		dispose();
		startServer();
	}

	private static synchronized void setupServer() throws Exception {
		AccessController.doPrivileged(new PrivilegedExceptionAction<Object>(){

			@Override
			public Object run() throws Exception {
				final NSFDataConnectionConfiguration nsfDataConnConfig=new NSFDataConnectionConfiguration();
				if (nsfDataConnConfig.isEnabled()) {
					FtpServerFactory serverFactory = new FtpServerFactory();

					ListenerFactory factory = new ListenerFactory();
					factory.setDataConnectionConfiguration(nsfDataConnConfig);

					// set the port of the listener
					m_port=nsfDataConnConfig.getPort();
					factory.setPort(m_port);

					// replace the default listener
					serverFactory.addListener("default", factory.createListener());
					//register user manager that reads from config db
					final NSFUserManager nsfUserManager=new NSFUserManager(nsfDataConnConfig);
					serverFactory.setUserManager(nsfUserManager);
					
					//read files from local server filesystem
					NativeFileSystemFactory fsFactory=new NativeFileSystemFactory();
					//auto-create home directories:
					fsFactory.setCreateHome(true);
					//case-insensitive on Windows:
					fsFactory.setCaseInsensitive("\\".equals(File.pathSeparator));
					serverFactory.setFileSystem(fsFactory);
					
					//prevent anonymous access if there is no user "Anonymous"
					DefaultConnectionConfig connConfig=new DefaultConnectionConfig() {
						@Override
						public boolean isAnonymousLoginEnabled() {
							return nsfUserManager.hasAnonymousUser();
						}
					};
					serverFactory.setConnectionConfig(connConfig);

					// start the server
					m_server = serverFactory.createServer(); 
					m_serverDisabled=false;
				}
				else {
					m_serverDisabled=true;
				}
				return null;
			}
		});
	}
}
