package com.mindoo.ftp.users.nsf;

import java.io.File;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Vector;
import java.util.logging.Level;

import lotus.domino.Database;
import lotus.domino.DateTime;
import lotus.domino.Document;
import lotus.domino.Item;
import lotus.domino.NotesException;
import lotus.domino.NotesFactory;
import lotus.domino.NotesThread;
import lotus.domino.Session;
import lotus.domino.View;
import lotus.domino.ViewEntry;
import lotus.domino.ViewNavigator;

import org.apache.ftpserver.ftplet.Authentication;
import org.apache.ftpserver.ftplet.AuthenticationFailedException;
import org.apache.ftpserver.ftplet.Authority;
import org.apache.ftpserver.ftplet.FtpException;
import org.apache.ftpserver.ftplet.User;
import org.apache.ftpserver.ftplet.UserManager;
import org.apache.ftpserver.usermanager.AnonymousAuthentication;
import org.apache.ftpserver.usermanager.UsernamePasswordAuthentication;
import org.apache.ftpserver.usermanager.impl.BaseUser;
import org.apache.ftpserver.usermanager.impl.ConcurrentLoginPermission;
import org.apache.ftpserver.usermanager.impl.TransferRatePermission;
import org.apache.ftpserver.usermanager.impl.UserMetadata;
import org.apache.ftpserver.usermanager.impl.WritePermission;

import com.mindoo.ftp.config.nsf.FTPDbLocator;
import com.mindoo.ftp.connection.nsf.NSFDataConnectionConfiguration;
import com.mindoo.ftp.internal.FTPActivator;

/**
 * Implementation of {@link UserManager} that reads the users from the NSF configuration database
 * 
 * @author Karsten Lehmann
 */
public class NSFUserManager implements UserManager {
	private NSFDataConnectionConfiguration m_nsfConfig;
	
	public NSFUserManager(NSFDataConnectionConfiguration nsfConfig) {
		m_nsfConfig=nsfConfig;
	}

	/**
	 * Method to check if a user "Anonymous" is configured
	 * 
	 * @return true if configured
	 */
	public boolean hasAnonymousUser() {
		try {
			return getUserByName("Anonymous")!=null;
		}
		catch (FtpException e) {
			FTPActivator.log(Level.SEVERE, "Could not check if user 'Anonymous' is configured", e);
			return false;
		}
	}
	
	@Override
	public User authenticate(Authentication authentication) throws AuthenticationFailedException {
		if (authentication instanceof AnonymousAuthentication) {
			AnonymousAuthentication anonymousAuth=(AnonymousAuthentication) authentication;
			UserMetadata userMetaData=anonymousAuth.getUserMetadata();

			User user;
			try {
				user = getUserByName("Anonymous");
			}
			catch (FtpException e1) {
				throw new AuthenticationFailedException(e1);
			}
			if (user==null) {
				if (userMetaData!=null) {
					InetAddress inetAddr=userMetaData.getInetAddress();
					//resolve hostname in inetAddr
					inetAddr.getHostName();
					FTPActivator.log(Level.WARNING, "Anonymous login request from "+inetAddr.toString()+" rejected. No user 'Anonymous' found.", null);
				}
				else {
					FTPActivator.log(Level.WARNING, "Anonymous login request rejected. No user 'Anonymous' found.", null);
				}
				throw new AuthenticationFailedException("No user 'Anonymous' found");
			}
			else {
				if (userMetaData!=null) {
					InetAddress inetAddr=userMetaData.getInetAddress();
					//resolve hostname in inetAddr
					inetAddr.getHostName();
					//make sure we have a fresh list of blocked networks
					m_nsfConfig.reloadBlockedNetworks();
					if (m_nsfConfig.isBlocked(inetAddr)) {
						FTPActivator.log(Level.INFO, "Anonymous login from "+inetAddr.toString()+" blocked. Address found in blocked IP/subnet list", null);
						throw new AuthenticationFailedException("Access denied");
					}
					FTPActivator.log(Level.INFO, "Anonymous login from "+inetAddr.toString()+" granted on directory "+user.getHomeDirectory(), null);
				}
				else {
					FTPActivator.log(Level.INFO, "Anonymous login granted on directory "+user.getHomeDirectory(), null);
				}
			}
			return user;
		}
		else if (authentication instanceof UsernamePasswordAuthentication) {
			UsernamePasswordAuthentication userPWAuth=(UsernamePasswordAuthentication) authentication;
			UserMetadata userMetaData=userPWAuth.getUserMetadata();
			String userName=userPWAuth.getUsername();
			String password=userPWAuth.getPassword();

			User user;
			try {
				user = getUserByName(userName);
			}
			catch (FtpException e1) {
				throw new AuthenticationFailedException(e1);
			}
			
			InetAddress inetAddr=userMetaData==null ? null : userMetaData.getInetAddress();
			if (inetAddr!=null) {
				//resolve hostname in inetAddr
				inetAddr.getHostName();
			}
			
			if (user==null) {
				if (inetAddr!=null) {
					FTPActivator.log(Level.WARNING, "Login request from "+inetAddr.toString()+" rejected. No user '"+userName+"' found.", null);
				}
				else {
					FTPActivator.log(Level.WARNING, "Login request rejected. No user '"+userName+"' found.", null);
				}
				throw new AuthenticationFailedException();
			}

			if (inetAddr!=null) {
				//make sure we read the most recent config (so that we can add blocked networks while the server is running)
				m_nsfConfig.reloadBlockedNetworks();
				//check if IP is blocked
				if (m_nsfConfig.isBlocked(inetAddr)) {
					FTPActivator.log(Level.INFO, "Login from "+inetAddr.toString()+" blocked. Address found in blocked IP/subnet list", null);
					throw new AuthenticationFailedException("Access denied");
				}
			}
			
			NotesThread.sinitThread();
			Session session=null;
			try {
				//let Notes verify the password for us; we don't store cleartext passwords
				session=NotesFactory.createSession();
				if (session.verifyPassword(password, user.getPassword())) {
					if (inetAddr!=null) {
						FTPActivator.log(Level.INFO, "Login from "+inetAddr.toString()+" and user "+userName+" granted on directory "+user.getHomeDirectory(), null);
					}
					else {
						FTPActivator.log(Level.INFO, "Login from user "+userName+" granted on directory "+user.getHomeDirectory(), null);
					}

					return user;
				}
				else {
					if (inetAddr!=null) {
						FTPActivator.log(Level.WARNING, "Login request from "+inetAddr.toString()+" rejected. Password incorrect.", null);
					}
					else {
						FTPActivator.log(Level.WARNING, "Login request rejected. Password incorrect.", null);
					}
					throw new AuthenticationFailedException();
				}
			}
			catch (NotesException e) {
				if (inetAddr!=null) {
					FTPActivator.log(Level.SEVERE, "Login request from "+inetAddr.toString()+" rejected, internal server error.", e);
				}
				else {
					FTPActivator.log(Level.SEVERE, "Login request rejected, internal server error.", e);
				}
				throw new AuthenticationFailedException(e);
			}
			finally {
				if (session!=null) {
					try {
						session.recycle();
					}
					catch (NotesException e) {
					}
				}
				NotesThread.stermThread();
			}
		}
		else {
			FTPActivator.log(Level.SEVERE, "Login request rejected, unsupported authentication method: "+authentication, null);
			throw new AuthenticationFailedException("No entry.");
		}
	}

	@Override
	public void delete(String login) throws FtpException {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean doesExist(String login) throws FtpException {
		return getUserByName(login)!=null;
	}

	@Override
	public String getAdminName() {
		return "admin";
	}

	@SuppressWarnings("unchecked")
	@Override
	public String[] getAllUserNames() throws FtpException {
		NotesThread.sinitThread();
		Session session=null;
		try {
			session=NotesFactory.createSession();
			Database configDb=FTPDbLocator.getConfigDatabase(session);
			if (configDb==null) {
				FTPActivator.log(Level.WARNING, "No FTP server configuration database found!", null);
				return null;
			}
			View lkView=configDb.getView("($lkUsersByLogin)");
			if (lkView==null) {
				FTPActivator.log(Level.WARNING, "View ($lkUsersByLogin) not found in FTP server configuration database", null);
				return null;
			}
			lkView.setAutoUpdate(false);
			ViewNavigator nav=lkView.createViewNav();
			ViewEntry currEntry=nav.getFirstDocument();
			if (currEntry!=null) {
				//improve viewnav performance
				nav.setCacheGuidance(400, ViewNavigator.VN_CACHEGUIDANCE_READSELECTIVE);
			}

			List<String> names=new ArrayList<String>();

			while (currEntry!=null) {
				if (currEntry.isValid()) {
					currEntry.setPreferJavaDates(true);
					//read first column value from view entries
					Vector colValues=currEntry.getColumnValues();
					Object firstColValue=colValues.get(0);
					if (firstColValue instanceof List) {
						List listValue=(List)firstColValue;
						if (!listValue.isEmpty())
							firstColValue=listValue.get(0);
					}
					if (firstColValue instanceof String) {
						names.add((String)firstColValue);
					}
				}
				ViewEntry nextEntry=nav.getNextDocument();
				currEntry.recycle();
				currEntry=nextEntry;
			}
			nav.recycle();
			return names.toArray(new String[names.size()]);
		}
		catch (NotesException e) {
			FTPActivator.log(Level.SEVERE, "Could not read user names from config db", e);
			throw new FtpException(e);
		}
		finally {
			if (session!=null) {
				try {
					session.recycle();
				}
				catch (NotesException e) {
					//
				}
			}
			NotesThread.stermThread();
		}
	}

	@Override
	public User getUserByName(String login) throws FtpException {
		NotesThread.sinitThread();
		Session session=null;
		try {
			session=NotesFactory.createSession();
			Database configDb=FTPDbLocator.getConfigDatabase(session);
			if (configDb==null) {
				FTPActivator.log(Level.WARNING, "No FTP server configuration database found!", null);
				return null;
			}
			View lkView=configDb.getView("($lkUsersByLogin)");
			if (lkView==null) {
				FTPActivator.log(Level.WARNING, "View ($lkUsersByLogin) not found in FTP server configuration database", null);
				return null;
			}

			//try to find user document
			Document userDoc=lkView.getDocumentByKey(login);
			if (userDoc==null) {
				return null;
			}
			else if (!userDoc.isValid()) {
				userDoc.recycle();
				return null;
			}

			//check if user account is enabled
			String enabled=userDoc.getItemValueString("Enabled");
			if (!"1".equals(enabled)) {
				userDoc.recycle();
				return null;
			}
			String timeControlledAccess=userDoc.getItemValueString("TimeControlledAccess");
			if ("1".equals(timeControlledAccess)) {
				Item fromItm=userDoc.getFirstItem("TimeControlledAccessFrom");
				Item untilItm=userDoc.getFirstItem("TimeControlledAccessUntil");
				if (fromItm==null || untilItm==null) {
					userDoc.recycle();
					return null;
				}
				
				DateTime ndtFrom=fromItm.getDateTimeValue();
				DateTime ndtUntil=untilItm.getDateTimeValue();
				Date now=new Date();
				
				if (ndtFrom!=null) {
					Date dtFrom=ndtFrom.toJavaDate();
					if (now.getTime() < dtFrom.getTime()) {
						//dtFrom is in the future
						userDoc.recycle();
						return null;
					}
				}
				if (ndtUntil!=null) {
					Date dtUntil=ndtUntil.toJavaDate();
					if (now.getTime() > dtUntil.getTime()) {
						//dtUntil is in the past
						userDoc.recycle();
						return null;
					}
				}
			}
			
			BaseUser newUser=new BaseUser();
			String password=userDoc.getItemValueString("Password");
			newUser.setPassword(password);
			String name=userDoc.getItemValueString("Name");
			newUser.setName(name);


			File mftpDirFile;
			String baseDirPath=session.getEnvironmentString("mndFTPBaseDir");
			if (baseDirPath==null || baseDirPath.length()==0) {
				//create directory <dominodatadir>/domino/mftp
				String dataDirPath=session.getEnvironmentString("Directory", true);
				File dataDirFile=new File(dataDirPath);
				File dominoDirFile=new File(dataDirFile, "domino");
				mftpDirFile=new File(dominoDirFile, "mftp");
				if (!mftpDirFile.mkdirs()) {
					//not allowed to create this directory; use system temp dir as last fallback
					File tmpFile=File.createTempFile("mftp", ".tmp");
					mftpDirFile=new File(tmpFile.getParentFile(), "mftp");
					if (!tmpFile.delete())
						tmpFile.deleteOnExit();
					mftpDirFile.mkdirs();
				}
			}
			else {
				mftpDirFile=new File(baseDirPath);
				mftpDirFile.mkdirs();
			}

			String subDirPath=userDoc.getItemValueString("Directory");
			if (File.separator.equals("\\")) {
				subDirPath=subDirPath.replace("/", "\\");
			}
			else {
				subDirPath=subDirPath.replace("\\", "/");
			}

			if (subDirPath.length()==0) {
				FTPActivator.log(Level.SEVERE, "Illegal directory configured for user "+login, null);
				throw new FtpException("Illegal directory configured for user "+login);
			}

			if (!subDirPath.endsWith(File.separator))
				subDirPath=subDirPath+File.separator;

			//prevent changing to parent directories
			subDirPath=subDirPath.replace("..", "").replace(":", "");
			
			//append subdirectory structure to base dir
			String homeDirPath=mftpDirFile.getAbsolutePath();
			if (!homeDirPath.endsWith(File.separator))
				homeDirPath=homeDirPath+File.separator;

			//if subdir path is empty
			if (subDirPath.startsWith(File.separator))
				subDirPath=subDirPath.substring(1);
			homeDirPath+=subDirPath;
			
			newUser.setHomeDirectory(homeDirPath);
			File homeDirFile=new File(homeDirPath);
			if (!homeDirFile.exists()) {
				//try to create the directory and log errors
				if (!homeDirFile.mkdirs()) {
					FTPActivator.log(Level.WARNING, "Could not create home directory '"+homeDirFile.getAbsolutePath()+"' for user "+login, null);
				}
			}
			boolean canWrite="1".equals(userDoc.getItemValueString("Perm_canWrite"));
			int concurrentLogins=userDoc.getItemValueInteger("Perm_ConcurrentLogins");
			int concurrentLoginsIP=userDoc.getItemValueInteger("Perm_ConcurrentLoginsIP");
			ConcurrentLoginPermission loginPerm=new ConcurrentLoginPermission(concurrentLogins, concurrentLoginsIP);

			int maxDownloadRate=userDoc.getItemValueInteger("Perm_MaxDownloadRate");
			int maxUploadRate=userDoc.getItemValueInteger("Perm_MaxUploadRate");
			TransferRatePermission transferPerm=new TransferRatePermission(maxDownloadRate, maxUploadRate);
			List<Authority> authorities=new ArrayList<Authority>();
			if (canWrite) {
				authorities.add(new WritePermission());
			}
			authorities.add(loginPerm);
			authorities.add(transferPerm);
			newUser.setAuthorities(authorities);
			newUser.setEnabled(true);
			userDoc.recycle();

			return newUser;
		}
		catch (Throwable e) {
			FTPActivator.log(Level.SEVERE, "Could not read user info from config db", e);
			throw new FtpException(e);
		}
		finally {
			if (session!=null) {
				try {
					session.recycle();
				}
				catch (NotesException e) {
					//
				}
			}
			NotesThread.stermThread();
		}
	}

	@Override
	public boolean isAdmin(String login) throws FtpException {
		return "admin".equals(login);
	}

	@Override
	public void save(User user) throws FtpException {
		throw new UnsupportedOperationException();
	}
}
